# A Constrained Bayesian Approach to Out-of-Distribution Prediction 

Here is the instruction for the DomainBed experiments (Section 6.2). 
Code and instructions for the synthetic experiments are in the `synth/` directory.

## Dependencies

Python 3.9, CUDA 11, PyTorch (for DomainBed), Jax.

For details about the DomainBed benchmark see https://github.com/facebookresearch/DomainBed/.

To set up the environment:
```
conda create -n drodg python==3.9 pip
source activate drodg
# if necessary, change `cu111` according to your CUDA version. 
# Also change the occurrences in `requirements.txt`, and for Jaxlib
pip install -r requirements.txt --index-url https://download.pytorch.org/whl/cu111
```

## Running the Experiments

Step 1. train the ConvNet feature extractors using DomainBed.

- For Modified ColoredMNIST:
```sh
python3 -m domainbed.scripts.sweep launch \
--data_dir=domainbed/data \
--output_dir sweep_output/ColoredMNIST01 \
--command_launcher multi_gpu \
--algorithm ERM \
--dataset ColoredMNIST01 \
--n_hparams 20 \
--n_trials 3 \
--hparams "{\"expand_factor\": 8, \"N_WORKERS\": 4}" \
--single_test_envs
```
- For PACS:
```sh
python -m domainbed.scripts.sweep launch \
--data_dir=domainbed/data \
--output_dir=sweep_output/PACS \
--command_launcher multi_gpu \
--algorithms ERM \
--datasets PACS \
--n_hparams 20 \
--n_trials 3 \
--hparams "{\"N_WORKERS\": 4}" \
--single_test_envs
```

Step 2. To reproduce Table 1, 2, S1 and S2 in the paper, run following commands 
        and you will get corresponding report files at 
        `{dataset}_N:{N}_M:{methods}_MI:{MI}}_REG:{REG}_GAP:{GAP}_txt`
```sh
# For CMNIST (Table 1 & S1):

for n_adapt in 4 8 16 32; do
    python jax_meta.py --eta 1 --dataset=ColoredMNIST01 --n_adapt $n_adapt --methods blr blc blc_n_train
done

for n_adapt in 4 8 16 32; do
    for GAP in 0.05 0.1 0.2; do
        python jax_meta.py --eta 1 --dataset=ColoredMNIST01 --n_adapt $n_adapt --GAP $GAP --methods cblr
    done
done

for n_adapt in 4 8 16 32; do
    for MI in 0.1 1 10; do 
        for REG in 0.1 1 10; do 
            python jax_meta.py --eta 1 --dataset=ColoredMNIST01 --n_adapt $n_adapt --MI $MI --REG $REG --methods divdis --divdis_init finetune
        done
    done
done

# For PACS (Table 2 & S2):

for n_adapt in 16 32 64 128 256; do
    python jax_meta.py --eta 50 --dataset=PACS --n_adapt $n_adapt --methods blr blc blc_n_train
done

for n_adapt in 16 32 64 128 256; do
    for GAP in 0.05 0.1 0.2; do
        python jax_meta.py --eta 50 --dataset=PACS --n_adapt $n_adapt --GAP $GAP --methods cblr
    done
done

for n_adapt in 16 32 64 128 256; do
    for MI in 0.1 1 10; do 
        for REG in 0.1 1 10; do 
            python jax_meta.py --eta 50 --dataset=PACS --n_adapt $n_adapt --MI $MI --REG $REG --methods divdis --divdis_init reinit
        done
    done
done
```

Alternatively, you may break down the above script:

a. Each directory in `sweep_output/{ColorMNIST01,PACS}` contains the NN weights 
for a single train/test domain split. The following script loads the 
corresponding weights in `${output_dir}` and save the NN embeddings in `embeddings` subdir.
```sh
python jax_save.py ${output_dir}
```

b. You can then run a single adaptation method (`${method}`) with the following:
```sh
python jax_train.py \
--output_dir ${output_dir} \  # XXX
--method ${method} \  # XXX
--steps 20000 \
--eta 1 \
--n_adapt 8 \  # XXX
--seed 0 \
```
where `${method}` can be one of 
- `cblr`: the proposed method
- `blr`: the BLR baseline
- `blc`: BLR-LC-N_adapt in paper
- `blc_n_train`: BLR-LC-N_train
- `divdis`: the DivDis baseline (Lee et al, 2022)

Or you can sweep over all methods for with the following:
```sh
python jax_sweep.py --n_adapt ${N} --eta 1 --steps 20000 --n_trial=20 --output_dir=${output_dir} --methods cblr blr blc blc_n_train divdis
```

c. Each of the Domainbed benchmarks has `n_envs` domains, so we use `jax_meta.py` 
to sweep over all methods and all train/test domain splits for a specified 
benchmark, and then generate a report in 
`{dataset}_N:{N}_M:{methods}_MI:{MI}}_REG:{REG}_GAP:{GAP}_txt`:
```sh
python jax_meta.py --n_adapt {N} --eta 1 --steps 20000 --n_trial=20 --dataset={dataset} --methods cblr blr blc blc_n_train divdis
```

