import apportionpy.apportionment as ap

import pickle
import ipdb
from pathlib import Path

import torch
import torch.nn as nn
import torch.nn.functional as F
import torch.autograd as autograd
import math
import numpy as np

import jax
import jax.numpy as jnp
from jax import grad, jit, vmap
from jax import random as random
from jax.config import config
config.update("jax_debug_nans", True)
config.update("jax_debug_infs", True)

from tqdm import trange


import argparse
import collections
import json
import os
import random as orandom
import sys
import time
import uuid

import numpy as np
import PIL
import torch
import torchvision
import torch.utils.data

from domainbed import datasets
from domainbed import hparams_registry
from domainbed import algorithms
from domainbed.lib import misc
from domainbed.lib.fast_data_loader import InfiniteDataLoader, FastDataLoader

from collections import defaultdict
from attrdict import AttrDict, AttrDefault

from domainbed.datasets import MultipleDomainDataset
from torch.utils.data import TensorDataset

import matplotlib.pyplot as plt


# +
# output_dir = 'sweep_output/ColoredMNIST/0d1869a41815fbc98e23f72469a1e3c5'
# output_dir = 'sweep_output/VLCS/2a5ff21f64b43ec900cce82cdea7794d'
# output_dir = 'train_output/PACS'

np.seterr(all='raise')

parser = argparse.ArgumentParser(description='Jax train in output_dir')
parser.add_argument('--output_dir', type=str, default='train_output/PACS')
parser.add_argument('--method', type=str, default='cblr')
parser.add_argument('--seed', type=int, default=0)
parser.add_argument('--load_from', type=int, default=0, help='from which step')
parser.add_argument('--steps', type=int, default=20000, help='to which step')
parser.add_argument('--checkpoint_freq', type=int, default=200)
parser.add_argument('--n_adapt', type=int, default=20)
parser.add_argument('--eta', type=float, default=100, help='ETA = eta * LR')
parser.add_argument('--debug', action="store_true")
parser.add_argument('--MI', type=float, default=10., help='mutual information')
parser.add_argument('--REG', type=float, default=1., help='regularizer')
parser.add_argument('--GAP', type=float, default=0.1, help='Acc lower bound = min source domain - GAP')
parser.add_argument('--divdis_init', type=str, default='reinit', choices=['reinit', 'finetune'], help='divdis heads initialization method')

local_args = parser.parse_args()
# local_args = parser.parse_args(args=[]) # for jupyter notebook




output_dir = local_args.output_dir
method = local_args.method
seed = local_args.seed
load_from = local_args.load_from

N_STEPS = local_args.steps
N_ADAPTATION = local_args.n_adapt  # number of adaption set per class
ETA_MULT = local_args.eta

N_MC = 10  # number of mc samples

jax_output_dir = f'{output_dir}/results_adapt:{N_ADAPTATION}_eta:{ETA_MULT}'.replace('.', '_')
os.makedirs(jax_output_dir, exist_ok=True)

save_dict = torch.load(os.path.join(output_dir, 'model.pkl'))
args = AttrDict(save_dict['args'])
hparams = AttrDict(save_dict['model_hparams'])

orandom.seed(args.seed)
np.random.seed(args.seed)
torch.manual_seed(args.seed)
torch.backends.cudnn.deterministic = True
torch.backends.cudnn.benchmark = False

if torch.cuda.is_available():
    device = "cuda"
else:
    device = "cpu"

if args.dataset in vars(datasets):
    dataset = vars(datasets)[args.dataset](args.data_dir,
                                           args.test_envs, hparams)
else:
    raise NotImplementedError

algorithm_class = algorithms.get_algorithm_class(args.algorithm)
algorithm = algorithm_class(dataset.input_shape, dataset.num_classes,
                            len(dataset) - len(args.test_envs), hparams)
algorithm.load_state_dict(save_dict["model_dict"])

if dataset.input_shape == (3, 224, 224):
    ETA = 5e-5 * ETA_MULT
else:
    ETA = 1e-3 * ETA_MULT



# +
embedding_ds = MultipleDomainDataset()
embedding_ds.datasets = []
embedding_ds.num_classes = dataset.num_classes
for i, d in enumerate(dataset.datasets):
    xs = torch.load(f"{args.output_dir}/embeddings/_DRODG_env_{i}_xs.pt")
    ys = torch.load(f"{args.output_dir}/embeddings/_DRODG_env_{i}_ys.pt")
    embedding_ds.datasets.append(TensorDataset(xs, ys))
    if local_args.debug:
        ys__ = ys.numpy().tolist()
        print("class distribution: ", [f"{ys__.count(ii) / len(ys__):.3g}" for ii in range(7)])
        print(f'{xs.shape=}, {ys.shape=}')
dataset = embedding_ds

# gaussian prior ~ N(0, \sigma^2) 
# NLL = \frac{1}{2 * \sigma^2} ||w||^2
# w_{t+1} = (1 - \frac{\eta}{\sigma^2}) w_t
# \eta \lt 2 * \sigma^2
d = xs.shape[1]
# \sigma^2 ~ ||w||^2 * (4~10) / d
sigma2 = 9. * 10 / d
WD = 0.5 / sigma2

# hparams['class_balanced'] = True
# If set true, samples in smaller class with be sampled with a proportinally larger prob

n_envs = len(dataset.datasets)
eval_names = [f'env{i}_in' for i in range(
    n_envs)] + [f'env{i}_out' for i in range(n_envs)]

# +
# according to the results. no need to clip embedding norm

# fig, axs = plt.subplots(1, len(dataset.datasets), sharey=True)
# n_bins=20
# for i, ds in enumerate(dataset.datasets):
#     xs = ds.tensors[0]
#     x_norms = (xs**2).sum(0).sqrt().numpy()
#     axs[i].hist(x_norms, bins=n_bins)

# +

def AttrDefault2Dict(attr_default):
    if hasattr(attr_default, 'keys'):
        return {k: AttrDefault2Dict(attr_default[k]) for k in attr_default}
    else:
        return attr_default


def torch2jax(tensor):
    # return jax.dlpack.from_dlpack(torch.utils.dlpack.to_dlpack(tensor))
    return jnp.array(tensor.detach().cpu().numpy())


def get_datas_from_splits(splits, frac=1.0):
    # return jax array
    datas = []
    for env, env_weights in splits:
        LEN = int(len(env) * frac)
        if hasattr(env, 'underlying_dataset'):
            data = env.underlying_dataset[:LEN]  # env[:LEN] causes error
        else:
            data = env[:LEN]
        datas.append((torch2jax(data[0]), torch2jax(data[1])))
    return datas


def draw_from_test_ds(test_ds, n_classes, n_adaptation, seed=0):
    # return torch tensor
    indices = [[] for i in range(n_classes)]
    for i, (_, y) in enumerate(test_ds):
        indices[y.item()].append(i)

    class_sizes = [len(indice) for indice in indices]
    class_n_adapts = ap.Apportion(seats=n_adaptation, populations=class_sizes, method='adam').fair_shares

    if local_args.debug:
        print(f'{class_sizes=}')
        print(f'{class_n_adapts=}')

    drawn_indices = []
    for i in range(n_classes):
        np.random.RandomState(seed).shuffle(indices[i])
        drawn_indices += indices[i][:class_n_adapts[i]]
#         drawn_indices += np.random.RandomState(seed).permutation(indices[i])[:class_n_adapts[i]]

    xs = torch.cat([test_ds[i][0].unsqueeze(0) for i in drawn_indices])
    ys = torch.cat([test_ds[i][1].unsqueeze(0) for i in drawn_indices])
    return xs, ys


def draw_from_splits(test_splits, seed=0):
    drawn_test_splits = []
    for env, _ in test_splits:
        drawn_data = draw_from_test_ds(
            env, dataset.num_classes, N_ADAPTATION, seed=seed)
        drawn_ds = TensorDataset(*drawn_data)
        w = misc.make_weights_for_balanced_classes(drawn_ds) if hparams['class_balanced'] else None
        drawn_test_splits.append((drawn_ds, w))
    return drawn_test_splits


# +
def get_splits():
    in_splits = []
    out_splits = []
    uda_splits = []
    for env_i, env in enumerate(dataset):
        uda = []

        out, in_ = misc.split_dataset(env,
                                      int(len(env)*args.holdout_fraction),
                                      misc.seed_hash(args.trial_seed, env_i))

        if env_i in args.test_envs:
            uda, in_ = misc.split_dataset(in_,
                                          int(len(in_)*args.uda_holdout_fraction),
                                          misc.seed_hash(args.trial_seed, env_i))

        if hparams['class_balanced']:
            in_weights = misc.make_weights_for_balanced_classes(in_)
            out_weights = misc.make_weights_for_balanced_classes(out)
            if uda is not None:
                uda_weights = misc.make_weights_for_balanced_classes(uda)
        else:
            in_weights, out_weights, uda_weights = None, None, None
        in_splits.append((in_, in_weights))
        out_splits.append((out, out_weights))
        if len(uda):
            uda_splits.append((uda, uda_weights))

    if args.task == "domain_adaptation" and len(uda_splits) == 0:
        raise ValueError("Not enough unlabeled samples for domain adaptation.")

    train_splits = [(env, env_weights) for i, (env, env_weights)
                    in enumerate(in_splits) if i not in args.test_envs]
    test_splits = [(env, env_weights) for i, (env, env_weights)
                   in enumerate(in_splits) if i in args.test_envs]
    all_splits = in_splits + out_splits + uda_splits

    return train_splits, test_splits, all_splits


def get_loaders(splits):
    loaders = [InfiniteDataLoader(
        dataset=env,
        weights=env_weights,
        batch_size=hparams['batch_size'],
        num_workers=hparams.N_WORKERS)
        for i, (env, env_weights) in enumerate(splits)]
    return loaders


train_splits, test_splits, all_splits = get_splits()
if local_args.debug:
    print(f'{len(train_splits)=}')
    print([len(split[0]) for split in train_splits])
    print(f'{len(test_splits)=}')
    print([len(split[0]) for split in test_splits])
    print(f'{len(all_splits)=}')
    print([len(split[0]) for split in all_splits])
train_datas = get_datas_from_splits(train_splits)
all_datas = get_datas_from_splits(all_splits)
if local_args.debug:
    print([(data[0].shape, data[1].shape) for data in train_datas])
    print([(data[0].shape, data[1].shape) for data in all_datas])


# +
def load_network_params(pytorch_MLP):
    w = torch2jax(pytorch_MLP.weight)
    b = torch2jax(pytorch_MLP.bias)
    return (w, b)


def jax_predict(params, x):
    w, b = params
    logits = jnp.dot(x, w.transpose()) + b
    return logits - jnp.expand_dims(jax.scipy.special.logsumexp(logits, axis=-1), -1)


def cross_entropy(logprobs, target_classes):
    nll = jnp.take_along_axis(
        logprobs, jnp.expand_dims(target_classes, axis=1), axis=1)
    ce = -jnp.sum(nll)
    return ce


def norm2(t):
    '''returns ||t||_2^2'''
    return sum(jax.tree_map(lambda a: (a**2).sum(), t))


def accuracy(params, data):
    xs, ys = data
    p = jax_predict(params, xs)
    correct = (p.argmax(-1) == ys).sum()
    total = p.shape[0]
    return correct / total


def chain_accuracy(cparams, data):
    xs, ys = data
    p = jnp.exp(jax.vmap(jax_predict, in_axes=(0, None))(cparams, xs)).mean(0)
    correct = (p.argmax(-1) == ys).sum()
    total = p.shape[0]
    return correct / total


def max_accuracy(cparams, data):
    accs = jax.vmap(accuracy, in_axes=(0 ,None))(cparams, data)
    return accs.max()


def _ece(p, ys, n_bins=20):

    p = np.array(p)
    ys = np.array(ys)
    if ys.ndim > 1:
        ys = np.argmax(ys, axis=-1)

    p_idx = np.argmax(p, axis=-1)
    p_val = np.max(p, axis=-1)
    p_val = np.array(p_val)
    acc, conf = np.zeros(n_bins), np.zeros(n_bins)
    Bm = np.zeros(n_bins)
    for m in range(n_bins):
        a, b = m / n_bins, (m+1) / n_bins
        for i in range(p.shape[0]):
            if p_val[i] > a and p_val[i] <= b:
                Bm[m] += 1
                if p_idx[i] == ys[i]:
                    acc[m] += 1
                conf[m] += p_val[i]
        if Bm[m] != 0:
            acc[m] = acc[m] / Bm[m]
            conf[m] = conf[m] / Bm[m]
    ece_total = 0
    for m in range(n_bins):
        ece_total += Bm[m] * np.abs(acc[m] - conf[m])
    return ece_total / sum(Bm)


def ece(params, data):
    xs, ys = data
    p = jnp.exp(jax_predict(params, xs))
    return _ece(p, ys)


def chain_ece(cparams, data):
    xs, ys = data
    p = jnp.exp(jax.vmap(jax_predict, in_axes=(0, None))(cparams, xs)).mean(0)
    return _ece(p, ys)


def loss(params, x, y):
    pred = jax_predict(params, x)
    return cross_entropy(pred, y) + WD * norm2(params)

def xent_loss(params, x, y):
    pred = jax_predict(params, x)
    return cross_entropy(pred, y)


# +
@jit
def update_cblr(params, x, y, key):
    grads = grad(loss)(params, x, y)
    THRESHOLD = ETA/(2**8)

    leaves, treedef = jax.tree_util.tree_flatten(params)
    key, nkey = random.split(key)
    nkeys = random.split(nkey, len(leaves))
    noises = [jax.random.normal(key, shape=leaf.shape)
              for leaf, key in zip(leaves, nkeys)]
    noises = jax.tree_util.tree_unflatten(treedef, noises)

    def sgld_step(eta):
        return jax.tree_map(lambda p, g, n: p - eta * g + jnp.sqrt(2 * eta) * n, params, grads, noises)

    def gd_step(eta, params, grads):
        return jax.tree_map(lambda p, g: p - eta * g, params, grads)

    def skip(eta):
        return params

    def any_bad(eta):
        cparams = sgld_step(eta)
        accs = []
        for data in train_datas:
            accs.append(accuracy(cparams, data))
        return jnp.min(jax.numpy.asarray(accs)) < MIN_ACC

    eta = jax.lax.while_loop(lambda eta: jnp.logical_and(
        eta > THRESHOLD, any_bad(eta)), lambda e: e / 2., ETA)
    
    params_1 = sgld_step(eta)
    grads_1 = grad(loss)(params_1, x, y)
    loss_0 = loss(params, x, y)
    loss_1 = loss(params_1, x, y)
    NLL_proposal_0 = 0.25 / eta * \
        norm2(jax.tree_map(lambda a, b: a-b, params,
              gd_step(eta, params_1, grads_1)))
    NLL_proposal_1 = 0.25 / eta * norm2(noises)
    accept_prob = jnp.minimum(1., jnp.exp(
        loss_0 - loss_1 - NLL_proposal_0 + NLL_proposal_1))

    return jax.lax.cond(eta > THRESHOLD, sgld_step, skip, eta), accept_prob


@jit
def update_crw(params, key):
    THRESHOLD = ETA/(2**8)

    leaves, treedef = jax.tree_util.tree_flatten(params)
    key, nkey = random.split(key)
    nkeys = random.split(nkey, len(leaves))
    noises = [jax.random.normal(key, shape=leaf.shape)
              for leaf, key in zip(leaves, nkeys)]
    noises = jax.tree_util.tree_unflatten(treedef, noises)

    def rw_step(eta):
        # WD as prior
        return jax.tree_map(lambda p, n: (1 - 2 * eta * WD) * p + jnp.sqrt(2 * eta) * n, params, noises)

    def skip(eta):
        return params

    def any_bad(eta):
        cparams = rw_step(eta)
        accs = []
        for data in train_datas:
            accs.append(accuracy(cparams, data))
        return jnp.min(jax.numpy.asarray(accs)) < MIN_ACC

    eta = jax.lax.while_loop(lambda eta: jnp.logical_and(
        eta > THRESHOLD, any_bad(eta)), lambda e: e / 2., ETA)
    return jax.lax.cond(eta > THRESHOLD, rw_step, skip, eta)


@jit
def update_clr(params, x, y, key):
    grads = grad(loss)(params, x, y)
    THRESHOLD = ETA/(2**8)

    def gd_step(eta):
        return jax.tree_map(lambda p, g: p - eta * g, params, grads)

    def skip(eta):
        return params

    def any_bad(eta):
        cparams = gd_step(eta)
        accs = []
        for data in train_datas:
            accs.append(accuracy(cparams, data))
        return jnp.min(jax.numpy.asarray(accs)) < MIN_ACC

    eta = jax.lax.while_loop(lambda eta: jnp.logical_and(
        eta > THRESHOLD, any_bad(eta)), lambda e: e / 2., ETA)
    return jax.lax.cond(eta > THRESHOLD, gd_step, skip, eta)


@jit
def update_blr(params, x, y, key):
    grads = grad(loss)(params, x, y)

    leaves, treedef = jax.tree_util.tree_flatten(params)
    key, nkey = random.split(key)
    nkeys = random.split(nkey, len(leaves))
    noises = [jax.random.normal(key, shape=leaf.shape)
              for leaf, key in zip(leaves, nkeys)]
    noises = jax.tree_util.tree_unflatten(treedef, noises)

    def sgld_step(eta, params, grads, noises):
        return jax.tree_map(lambda p, g, n: p - eta * g + jnp.sqrt(2 * eta) * n, params, grads, noises)

    def gd_step(eta, params, grads):
        return jax.tree_map(lambda p, g: p - eta * g, params, grads)

    params_1 = sgld_step(ETA, params, grads, noises)
    grads_1 = grad(loss)(params_1, x, y)

    loss_0 = loss(params, x, y)
    loss_1 = loss(params_1, x, y)
    NLL_proposal_0 = 0.25 / ETA * \
        norm2(jax.tree_map(lambda a, b: a-b, params,
              gd_step(ETA, params_1, grads_1)))
    NLL_proposal_1 = 0.25 / ETA * norm2(noises)
    accept_prob = jnp.minimum(1., jnp.exp(
        loss_0 - loss_1 - NLL_proposal_0 + NLL_proposal_1))

    return sgld_step(ETA, params, grads, noises) , accept_prob

@jit
def update_blc_n_train(params, x_train, y_train, x_adapt, y_adapt, n_train, n_adapt, key):
    grads_train = grad(loss)(params, x_train, y_train)
    grads_adapt = grad(loss)(params, x_adapt, y_adapt)
    alpha = 0.02 # N := alpha * N_train
    # grads = jax.tree_map(lambda a, b: a * alpha * n_train / (n_envs-1) + b * n_adapt, grads_train, grads_adapt)
    # eta = ETA / (alpha * n_train + n_adapt) * n_envs
    grads = jax.tree_map(lambda a, b: a * alpha * n_train * n_envs / (n_envs-1) / (alpha * n_train + n_adapt) + b * n_adapt * n_envs / (alpha * n_train + n_adapt), grads_train, grads_adapt)
    eta = ETA 

    leaves, treedef = jax.tree_util.tree_flatten(params)
    key, nkey = random.split(key)
    nkeys = random.split(nkey, len(leaves))
    noises = [jax.random.normal(key, shape=leaf.shape)
              for leaf, key in zip(leaves, nkeys)]
    noises = jax.tree_util.tree_unflatten(treedef, noises)

    def sgld_step(eta, params, grads, noises):
        return jax.tree_map(lambda p, g, n: p - eta * g + jnp.sqrt(2 * eta) * n, params, grads, noises)

    return sgld_step(eta, params, grads, noises) 


@jit
def update_lr(params, x, y, key):
    grads = grad(loss)(params, x, y)
    return jax.tree_map(lambda p, g: p - ETA/2**4 * g, params, grads)


# +
params = load_network_params(algorithm.classifier)

train_accs = [accuracy(params, data).item() for data in train_datas]
# MIN_ACC = min(train_accs) - 0.1
MIN_ACC = min(train_accs) - local_args.GAP # For looser acc bound

if local_args.debug: 
    print(f'{train_accs=}')
    print(f'{MIN_ACC=}')









def train_crw(params, seed=0, load_from=0):
    key = random.PRNGKey(seed)
    chain_params = jax.tree_map(
        lambda p: jnp.expand_dims(p, 0).repeat(N_MC, 0), params)
    cparams_crw = chain_params

    record = AttrDefault(list)
    if load_from != 0:
        cparams_crw, record = load_ckpt(method='crw', seed=seed, step=load_from)

    for step in trange(load_from, N_STEPS):

        if (step+1) % local_args.checkpoint_freq == 0 or step == 0:
            for name, data in zip(eval_names, all_datas):
                acc = chain_accuracy(cparams_crw, data).item()
                record[f'{name}_acc'].append(acc)
                ece_ = chain_ece(cparams_crw, data).item()
                record[f'{name}_ece'].append(ece_)
                if local_args.debug: print(f'{name=}:{acc=}:{ece_=}')
            if local_args.debug: print(f'weight norm={jnp.mean(jax.vmap(norm2)(cparams_crw)).item()}')
            save_ckpt(cparams_crw, record, method='crw', seed=seed, step=step)

        key, c_key = jax.random.split(key)
        c_keys = jax.random.split(c_key, N_MC)
        cparams_crw = jax.vmap(update_crw, in_axes=(0, 0))(
            cparams_crw, c_keys)

        

    return cparams_crw, record


def train_cblr(params, seed=0, load_from=0):
    key = random.PRNGKey(seed)
    chain_params = jax.tree_map(
        lambda p: jnp.expand_dims(p, 0).repeat(N_MC, 0), params)
    cparams_cblr = chain_params

    drawn_test_splits = draw_from_splits(test_splits, seed=seed)
    drawn_test_datas = get_datas_from_splits(drawn_test_splits)

    jax_x, jax_y = jnp.concatenate(list(
        zip(*drawn_test_datas))[0]), jnp.concatenate(list(zip(*drawn_test_datas))[1])

    record = AttrDefault(list)
    if load_from != 0:
        cparams_cblr, record = load_ckpt(method='cblr', seed=seed, step=load_from)

    for step in trange(load_from, N_STEPS):

        if (step+1) % local_args.checkpoint_freq == 0 or step == 0:
            for name, data in zip(eval_names, all_datas):
                acc = chain_accuracy(cparams_cblr, data).item()
                record[f'{name}_acc'].append(acc)
                ece_ = chain_ece(cparams_cblr, data).item()
                record[f'{name}_ece'].append(ece_)
                if local_args.debug: print(f'{name=}:{acc=}:{ece_=}')
            if local_args.debug: print(f'weight norm={jnp.mean(jax.vmap(norm2)(cparams_cblr)).item()}')
            save_ckpt(cparams_cblr, record, method='cblr', seed=seed, step=step)

        key, c_key = jax.random.split(key)
        c_keys = jax.random.split(c_key, N_MC)
        cparams_cblr, accept_probs = jax.vmap(update_cblr, in_axes=(0, None, None, 0))(
            cparams_cblr, jax_x, jax_y, c_keys)
        # accept_prob = accept_probs.mean().item()
        # record['accept_prob'].append(accept_prob)
        # if local_args.debug: print(f'{accept_prob=}')

        

    return cparams_cblr, record


def train_clr(params, seed=0, load_from=0):
    key = random.PRNGKey(seed)
    params_clr = params

    drawn_test_splits = draw_from_splits(test_splits, seed=seed)
    drawn_test_datas = get_datas_from_splits(drawn_test_splits)
    jax_x, jax_y = jnp.concatenate(list(
        zip(*drawn_test_datas))[0]), jnp.concatenate(list(zip(*drawn_test_datas))[1])

    record = AttrDefault(list)
    if load_from != 0:
        params_clr, record = load_ckpt(method='clr', seed=seed, step=load_from)
    for step in trange(load_from, N_STEPS):

        if (step+1) % local_args.checkpoint_freq == 0 or step == 0:
            for name, data in zip(eval_names, all_datas):
                acc = accuracy(params_clr, data).item()
                record[f'{name}_acc'].append(acc)
                ece_ = ece(params_clr, data).item()
                record[f'{name}_ece'].append(ece_)
                
            save_ckpt(params_clr, record, method='clr', seed=seed, step=step)

        params_clr = update_clr(params_clr, jax_x, jax_y, key)

        
    return params_clr, record


def train_blr(params, seed=0, load_from=0):
    key = random.PRNGKey(seed)
    chain_params = jax.tree_map(
        lambda p: jnp.expand_dims(p, 0).repeat(N_MC, 0), params)
    cparams_blr = chain_params

    drawn_test_splits = draw_from_splits(test_splits, seed=seed)
    drawn_test_datas = get_datas_from_splits(drawn_test_splits)
    jax_x, jax_y = jnp.concatenate(list(
        zip(*drawn_test_datas))[0]), jnp.concatenate(list(zip(*drawn_test_datas))[1])

    record = AttrDefault(list)
    if load_from != 0:
        cparams_blr, record = load_ckpt(method='blr', seed=seed, step=load_from)
    for step in trange(load_from, N_STEPS):

        if (step+1) % local_args.checkpoint_freq == 0 or step == 0:
            for name, data in zip(eval_names, all_datas):
                acc = chain_accuracy(cparams_blr, data).item()
                record[f'{name}_acc'].append(acc)
                ece_ = chain_ece(cparams_blr, data).item()
                record[f'{name}_ece'].append(ece_)
                if local_args.debug: print(f'{name=}:{acc=}:{ece_=}')
            if local_args.debug: print(f'weight norm={jnp.mean(jax.vmap(norm2)(cparams_blr)).item()}')
            save_ckpt(cparams_blr, record, method='blr', seed=seed, step=step)

        key, c_key = jax.random.split(key)
        c_keys = jax.random.split(c_key, N_MC)
        cparams_blr, accept_probs = jax.vmap(update_blr, in_axes=(
            0, None, None, 0))(cparams_blr, jax_x, jax_y, c_keys)
        accept_prob = accept_probs.mean().item()

        
    return cparams_blr, record


def train_blc(params, seed=0, load_from=0):
    key = random.PRNGKey(seed)
    chain_params = jax.tree_map(
        lambda p: jnp.expand_dims(p, 0).repeat(N_MC, 0), params)
    cparams_blc = chain_params

    drawn_test_splits = draw_from_splits(test_splits, seed=seed)
    in_loaders = get_loaders(train_splits + drawn_test_splits)
    if local_args.debug:
        print("train_splits: ", [len(split) for split, _ in train_splits])
        print("drawn_test_splits: ", [len(split) for split, _ in drawn_test_splits])

    record = AttrDefault(list)
    if load_from != 0:
        cparams_blc, record = load_ckpt(method='blc', seed=seed, step=load_from)
    for step in trange(load_from, N_STEPS):

        if (step+1) % local_args.checkpoint_freq == 0 or step == 0:
            for name, data in zip(eval_names, all_datas):
                acc = chain_accuracy(cparams_blc, data).item()
                record[f'{name}_acc'].append(acc)
                ece_ = chain_ece(cparams_blc, data).item()
                record[f'{name}_ece'].append(ece_)
            save_ckpt(cparams_blc, record, method='blc', seed=seed, step=step)

        minibatches = [(x.to(device), y.to(device))
                       for x, y in next(zip(*in_loaders))]
        all_x = torch.cat([x for x, y in minibatches])
        all_y = torch.cat([y for x, y in minibatches])
        jax_x = torch2jax(all_x)
        jax_y = torch2jax(all_y)

        key, c_key = jax.random.split(key)
        c_keys = jax.random.split(c_key, N_MC)
        # here blc using the same update_blr function
        cparams_blc, _ = jax.vmap(update_blr, in_axes=(
            0, None, None, 0))(cparams_blc, jax_x, jax_y, c_keys)

        
    return cparams_blc, record

from einops import rearrange

EPSILON = 1e-6
LAMBDA_MI = local_args.MI
LAMBDA_REG = local_args.REG

def kl_divergence(p, q):
  # p and q are JAX arrays of probabilities that sum to 1
  return jnp.sum(p * jnp.log(p / (q + EPSILON) + EPSILON))

def loss_divdis(cparams_dd, x_s, y_s, x_t, y_t, key):
    

    loss_XENT = jax.vmap(xent_loss, in_axes=(0, None, None))(cparams_dd, x_s, y_s).mean()

    probs = jax.vmap(jax_predict, in_axes=(0, None))(cparams_dd, x_t)
    probs = jnp.einsum("hbd->bhd", probs)
    probs = jax.nn.softmax(probs, axis=-1)
    B, H, D = probs.shape # B=batch_size, H=heads, D=pred_dim
    marginal_p = probs.mean(0) # H, D
    marginal_p = jnp.einsum("hd,ge->hgde", marginal_p, marginal_p) # H, H, D, D
    marginal_p = rearrange(marginal_p, "h g d e -> (h g) (d e)") # H^2, D^2
    joint_p = jnp.einsum("bhd,bge->bhgde", probs, probs).mean(0) # H, H, D, D
    joint_p = rearrange(joint_p, "h g d e -> (h g) (d e)") # H^2, D^2
    kl_divs = joint_p * (jnp.log(joint_p + EPSILON) - jnp.log(marginal_p + EPSILON))
    kl_grid = rearrange(kl_divs.sum(-1), "(h g) -> h g", h=H) # H, H
    pairwise_mis = jnp.triu(kl_grid, k=1) # Get only off-diagonal KL divergences
    loss_MI = pairwise_mis.mean()

    probs_mean_source = jax.nn.one_hot(y_s, num_classes=dataset.num_classes).mean(0)
    probs_mean_targets = probs.mean(0)
    loss_REG = jax.vmap(kl_divergence, in_axes=(0, None))(probs_mean_targets, probs_mean_source).mean()

    if local_args.debug:
        print(f'{LAMBDA_MI=}')
        print(f'{LAMBDA_REG=}')
        jax.debug.print("loss_XENT: {}", loss_XENT)
        jax.debug.print("loss_MI: {}", loss_MI)
        jax.debug.print("loss_REG: {}", loss_REG)

    # return loss_XENT 
    return loss_XENT + LAMBDA_MI * loss_MI + LAMBDA_REG * loss_REG



def train_divdis(params, seed=0, load_from=0):
    key = random.PRNGKey(seed)

    if local_args.divdis_init == 'reinit':
        # reinitialize N heads
        key, w_key = jax.random.split(key)
        key, b_key = jax.random.split(key)
        cparams_dd = jax.random.normal(w_key, (N_MC, *params[0].shape)), jax.random.normal(b_key, (N_MC, *params[1].shape)),
        cparams_dd = jax.tree_map(lambda p: p * 0.01, cparams_dd)
    elif local_args.divdis_init == 'finetune':
        # finetune init
        key, w_key = jax.random.split(key)
        key, b_key = jax.random.split(key)
        cparams_dd = jax.random.normal(w_key, (N_MC, *params[0].shape)), jax.random.normal(b_key, (N_MC, *params[1].shape)),
        cparams_dd = jax.tree_map(lambda p: p * 0.01, cparams_dd)
        cparams_dd = jax.tree_map(lambda cp, p: cp + jnp.expand_dims(p, 0).repeat(N_MC, 0), cparams_dd, params)
    else:
        raise

    drawn_test_splits = draw_from_splits(test_splits, seed=seed)
    n_train = sum([len(split) for split, _ in train_splits])
    n_adapt = sum([len(split) for split, _ in drawn_test_splits])
    if local_args.debug:
        print(f'{seed=}')
        print("train_splits: ", [len(split) for split, _ in train_splits])
        print("drawn_test_splits: ", [len(split) for split, _ in drawn_test_splits])
        print(f"{n_train=}, {n_adapt=}")
    train_loaders = get_loaders(train_splits)
    adapt_loaders = get_loaders(drawn_test_splits)

    drawn_test_splits = draw_from_splits(test_splits, seed=seed)
    drawn_test_datas = get_datas_from_splits(drawn_test_splits)

    jax_x, jax_y = jnp.concatenate(list(
        zip(*drawn_test_datas))[0]), jnp.concatenate(list(zip(*drawn_test_datas))[1])
    
    record = AttrDefault(list)
    for step in trange(0, N_STEPS):

       
        minibatches_train = [(x.to(device), y.to(device)) for x, y in next(zip(*train_loaders))]
        minibatches_adapt = [(x.to(device), y.to(device)) for x, y in next(zip(*adapt_loaders))]
        x_train = torch2jax(torch.cat([x for x, y in minibatches_train]))
        y_train = torch2jax(torch.cat([y for x, y in minibatches_train]))
        x_adapt = torch2jax(torch.cat([x for x, y in minibatches_adapt]))
        y_adapt = torch2jax(torch.cat([y for x, y in minibatches_adapt]))

        # diversify
        key, c_key = jax.random.split(key)
        g_cparams_dd = grad(loss_divdis)(cparams_dd, x_train, y_train, x_adapt, y_adapt, c_key)
        cparams_dd = jax.tree_map(lambda p, g: p - ETA * g, cparams_dd, g_cparams_dd)
        # disambiguate  
        accs = jax.vmap(accuracy, in_axes=(0 ,None))(cparams_dd, (jax_x, jax_y))
        idx = jnp.argmax(accs).item()
        # params_dd = cparams_dd[0][idx], cparams_dd[1][idx], 
        params_dd = jax.tree_map(lambda p: p[idx], cparams_dd)


        if (step+1) % local_args.checkpoint_freq == 0 or step == 0:

            if local_args.debug: 
                print(f'{idx=}')
                print(accs)
                print(f'weight norm={jnp.mean(jax.vmap(norm2)(cparams_dd)).item()}')
                print(f'grad norm={jnp.mean(jax.vmap(norm2)(g_cparams_dd)).item()}')

            for name, data in zip(eval_names, all_datas):
                acc = accuracy(params_dd, data).item()
                record[f'{name}_acc'].append(acc)
                # ece_ = ece(params_dd, data).item()
                # record[f'{name}_ece'].append(ece_)
                if local_args.debug: print(f'{name=}:{acc=}')
            save_ckpt(params_dd, record, method='divdis', seed=seed, step=step)
    return params_dd, record


def train_blc_n_train(params, seed=0, load_from=0):
    key = random.PRNGKey(seed)
    chain_params = jax.tree_map(
        lambda p: jnp.expand_dims(p, 0).repeat(N_MC, 0), params)
    cparams_blc = chain_params

    drawn_test_splits = draw_from_splits(test_splits, seed=seed)
    n_train = sum([len(split) for split, _ in train_splits])
    n_adapt = sum([len(split) for split, _ in drawn_test_splits])
    if local_args.debug:
        print("train_splits: ", [len(split) for split, _ in train_splits])
        print("drawn_test_splits: ", [len(split) for split, _ in drawn_test_splits])
        print(f"{n_train=}, {n_adapt=}")
    train_loaders = get_loaders(train_splits)
    adapt_loaders = get_loaders(drawn_test_splits)

    record = AttrDefault(list)
    if load_from != 0:
        cparams_blc, record = load_ckpt(method='blc_n_train', seed=seed, step=load_from)
    for step in trange(load_from, N_STEPS):

        if (step+1) % local_args.checkpoint_freq == 0 or step == 0:
            for name, data in zip(eval_names, all_datas):
                acc = chain_accuracy(cparams_blc, data).item()
                record[f'{name}_acc'].append(acc)
                ece_ = chain_ece(cparams_blc, data).item()
                record[f'{name}_ece'].append(ece_)
                if local_args.debug: print(f'{name=}:{acc=}:{ece_=}')
            if local_args.debug: print(f'weight norm={jnp.mean(jax.vmap(norm2)(cparams_blc)).item()}')
            save_ckpt(cparams_blc, record, method='blc_n_train', seed=seed, step=step)

        minibatches_train = [(x.to(device), y.to(device)) for x, y in next(zip(*train_loaders))]
        minibatches_adapt = [(x.to(device), y.to(device)) for x, y in next(zip(*adapt_loaders))]
        x_train = torch2jax(torch.cat([x for x, y in minibatches_train]))
        y_train = torch2jax(torch.cat([y for x, y in minibatches_train]))
        x_adapt = torch2jax(torch.cat([x for x, y in minibatches_adapt]))
        y_adapt = torch2jax(torch.cat([y for x, y in minibatches_adapt]))

        key, c_key = jax.random.split(key)
        c_keys = jax.random.split(c_key, N_MC)
        cparams_blc = jax.vmap(update_blc_n_train, in_axes=(
            0, None, None, None, None, None, None, 0))(cparams_blc, x_train, y_train, x_adapt, y_adapt, n_train, n_adapt, c_keys)
    return cparams_blc, record


def train_lc(params, seed=0, load_from=0):
    key = random.PRNGKey(seed)
    params_lc = params

    drawn_test_splits = draw_from_splits(test_splits, seed=seed)
    in_loaders = get_loaders(train_splits + drawn_test_splits)

    record = AttrDefault(list)

    if load_from != 0:
        params_lc, record = load_ckpt(method='lc', seed=seed, step=load_from)

    for step in trange(load_from, N_STEPS):

        if (step+1) % local_args.checkpoint_freq == 0 or step == 0:
            for name, data in zip(eval_names, all_datas):
                acc = accuracy(params_lc, data).item()
                record[f'{name}_acc'].append(acc)
                ece_ = ece(params_lc, data).item()
                record[f'{name}_ece'].append(ece_)
            save_ckpt(params_lc, record, method='lc', seed=seed, step=step)

        minibatches = [(x.to(device), y.to(device))
                    for x, y in next(zip(*in_loaders))]
        all_x = torch.cat([x for x, y in minibatches])
        all_y = torch.cat([y for x, y in minibatches])
        jax_x = torch2jax(all_x)
        jax_y = torch2jax(all_y)

        key, c_key = jax.random.split(key)
        c_keys = jax.random.split(c_key, 1)
        params_lc = update_lr(params_lc, jax_x, jax_y, key)

        

    return params_lc, record
# +


def save_params(params, path):
    params = jax.device_get(params)
    with open(path, 'wb+') as fp:
        pickle.dump(params, fp)


def load_params(path):
    with open(path, 'rb') as fp:
        params = pickle.load(fp)
    return jax.device_put(params)

def save_ckpt(trained_params, train_record, method, seed, step):
    torch.save(AttrDefault2Dict(train_record),
               f'{jax_output_dir}/{method}_trial{seed}_record.pt')
    save_params(trained_params,
                f'{jax_output_dir}/{method}_trial{seed}_params.npy')

def load_ckpt(method, seed, step):
    record = torch.load(f'{jax_output_dir}/{method}_trial{seed}_record.pt')
    record = AttrDefault(list, record)
    params = load_params(f'{jax_output_dir}/{method}_trial{seed}_params.npy')
    return params, record


    

if __name__ == '__main__':
    print(f'{method=}:{seed=}')
    trained_params, train_record = globals(
    )[f'train_{method}'](params, seed, load_from)



