This directory holds the code for reproducing the synthetic experiments (Section 6.1).

## Dependencies

You should be able to reuse the environment for the DomainBed experiment. 
Alternatively you can use the following:
```
conda create -n drodg python==3.9 pip
source activate drodg
pip install -r requirements.txt
```
Note that the version of `jaxlib` in `requirements.txt` only works with CUDA 11 and CuDNN 8.2. 
If you are using other versions, you need to change the corresponding line in `requirements.txt` 
to the URL of some other variants of jaxlib **0.4.7**, listed in  
[https://storage.googleapis.com/jax-releases/jax_cuda_releases.html](https://storage.googleapis.com/jax-releases/jax_cuda_releases.html).

## Running the Code

To reproduce Fig 1, S1 and S2 in the paper:
```
mkdir -p ~/run/drodg
source activate drodg
cd scripts
python classification-main-and-12k.py -ng ${NUMBER_OF_GPUS} -nm 2 -ns 32 -train  # use -nm 1 if you have GPU memory issues 
python regression-main-and-12k.py -ng ${NUMBER_OF_GPUS} -nm 2 -train
```
You will have two directories with names similar to `~/run/drodg/MM-DD-hh-mm_classification-main-and-12k_`.
You can then generate the figures using 
```
python plot_sim.py --classif_path ${PATH_FOR_CLASSIFICATION_EXPERIMENT} --regr_path ${PATH_FOR_REGRESSION_EXPERIMENT} --fig_number 1
```
You can change `fig_number` to `S1` or `S2` for the respective figure, and skip 
the `--regr_path` argument if you didn't run that experiment.

## NOTE

* The code will only run if it lives in a git repo (`git rev-parse HEAD` doesn't fail).

* the output will not exactly match the figures in paper, because in these figures 
the *true* predictor parameters are randomly sampled *only once* (and the error bars only 
correspond to the randomness in initializing and updating the MCMC particles). But they should
be qualitatively similar; `scripts/sample_figs` contains the plots across 3 sets of 
independent runs, which provide some ideas for the variation.

* The full sets of experiments will take 10-20 hours on 6 GPUs.  You can reduce the time cost 
by reducing the number of seeds (`-ns` for classification) and skip the regression experiment. 
