"""
initial scan for sensible range of (problem, algorithm) hyperparameters
"""

import monad_do as M
from experiments.master import runner
from experiments.utils import _get_timestr
import os
import shutil
import sys
import argparse

try:
    from .utils import *
except:
    from utils import *

parser = argparse.ArgumentParser()
parser.add_argument('--n_max_gpus', '-ng', type=int, default=2)
parser.add_argument('--n_multiplex', '-nm', type=int, default=2)
parser.add_argument('--n_seeds', '-ns', type=int, default=32)
parser.add_argument('--n_designs', '-nds', type=int, default=5)
parser.add_argument('-rmk', type=str, default='')
parser.add_argument('-train', action='store_true', default=False)


args = parser.parse_args()

@M.do(M.List)
def list_hps(__info):
    mu_a = 1  # yield [1, 0, 0.5**0.5]
    mu_b = (1 - mu_a**2) ** 0.5
    n_train = yield [12000, 6000]
    n_iters = 2500
    grid_size = 8
    n_adapt_max = 256
    dist_seed = yield list(range(args.n_designs))
    main_seed = yield list(range(dist_seed*50, dist_seed*50+args.n_seeds))

    return [proc(locals())]


remark = os.path.basename(__file__).split('.')[0] + '_' + args.rmk
exp_dir_base = os.path.expanduser(f'~/run/drodg')
log_dir_base = os.path.join(exp_dir_base, f'{_get_timestr()}_{remark}')
env_pref = f'CUDA_DEVICE_ORDER=PCI_BUS_ID XLA_PYTHON_CLIENT_MEM_FRACTION={0.9/args.n_multiplex:.3f} OMP_NUM_THREADS=2 '
root_cmd = env_pref + 'python main.py -production --constraint_errs 0.05 0.1 0.2 0.4'
tasks = list_hps(Info(root_cmd=root_cmd, log_dir_base=log_dir_base))


print('\n'.join(t.cmd for t in tasks[-1000:]))
print(len(tasks))
assert not os.system('df -h '+exp_dir_base)
if not args.train:
    sys.exit(0)

os.makedirs(log_dir_base, exist_ok=True)
shutil.copyfile(__file__, os.path.join(log_dir_base, 'script.py'))
with open(os.path.join(log_dir_base, 'script.py'), 'a') as fout:
    print('#', ' '.join(sys.argv), file=fout)

with open(os.path.join(log_dir_base, 'NAME'), 'w') as fout:
    print(args.rmk, file=fout)

r = runner.Runner(
    n_max_gpus=args.n_max_gpus, n_multiplex=args.n_multiplex, n_max_retry=-1)
r.run_tasks(tasks)

with open(os.path.join(log_dir_base, 'COMPLETE'), 'w') as fout:
    fout.write('.')
# init.py -ng 6 -nm 2 -rmk classif-addfull-12k -train
